/*
 *
 * Copyright (c) 2016-2018 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Author: Rafal Pietruch <r.pietruch@samsung.com>
 */
/**
 * @file crash-stack-libunw.c
 * @brief unwinding call stacks, functions specific for archs that use only libunwind
 */
#include <libunwind.h>
#include <limits.h>
#include <string.h>

#include "crash-stack.h"

#define LOG_TAG "CRASH_STACK"
#include "shared/log.h"

#define MAXPROCNAMELEN 1024

void _create_crash_stack(Dwfl *dwfl, pid_t pid, Callstack *callstack, elf_gregset_t *regs)
{
	// reimplemented based on libunwind tests/test-ptrace.c file
	unw_addr_space_t as = 0;
	void *ui = 0;
	do {
		//init before return
		callstack->elems = 0;

		as = unw_create_addr_space(_TB_accessors(), 0);
		if (!as)
			break;

		ui = _TB_create(pid);
		if (!ui)
			break;

		unw_cursor_t cursor;
		if (unw_init_remote(&cursor, as, ui) < 0)
			break;

		char proc_name[MAXPROCNAMELEN];
		char mod_name[PATH_MAX];
		for (; callstack->elems < sizeof(callstack->proc)/sizeof(callstack->proc[0]); ) {

			unw_word_t ip;
			if (unw_get_reg(&cursor, UNW_REG_IP, &ip) < 0)
				break;
			callstack->proc[callstack->elems].addr = ip;

			proc_name[0] = '\0';
			mod_name[0] = '\0';
			unw_word_t off;
			unw_word_t off_m;
			if (unw_get_proc_name(&cursor, proc_name, sizeof(proc_name), &off) == 0) {
				callstack->proc[callstack->elems].offset = off;
				callstack->proc[callstack->elems].name = strdup(proc_name);
			} else {
				callstack->proc[callstack->elems].offset = 0;
				callstack->proc[callstack->elems].name = NULL;
			}
			if (get_mod_name(ip, mod_name, PATH_MAX, &off_m) == 0) {
				callstack->proc[callstack->elems].module_offset = off_m;
				callstack->proc[callstack->elems].module_name = strdup(mod_name);
			} else {
				callstack->proc[callstack->elems].module_offset = 0;
				callstack->proc[callstack->elems].module_name = NULL;
			}
			++callstack->elems;

			if (unw_step(&cursor) <= 0)
				break;
		}
	} while (0);

	if (ui)
		_TB_destroy(ui);
	if (as)
		unw_destroy_addr_space(as);
}
