/*
 * Copyright (c) 2016,2018 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors: Adrian Szyndela <adrian.s@samsung.com>
 *          Łukasz Stelmach <l.stelmach@samsung.com>
 */
/**
 * @file crash-stack-x86_64.c
 * @brief unwinding call stacks, functions specific for x86_64
 */
#include "crash-stack.h"
#include <sys/user.h>
#include <string.h>

struct user_regs_struct g_registers;	///< static storage for registers

void *_crash_stack_get_memory_for_registers(size_t *size)
{
	if (NULL != size)
		*size = sizeof(g_registers);
	return &g_registers;
}

void *_get_place_for_register_value(const char *regname, int regnum)
{
	/* Function is not used for x86 anymore, make it dummy
	 * It is still called by generic __get_registers_core function
	 * The return value is checked for NULL in __get_registers_core.
	 */
	return NULL;
}

void _crash_stack_set_registers(void *regbuf)
{
	/* Make it dummy as it not used for x86 anymore
	 * However it is still called by generic __get_registers functions
	 */
}

unsigned long _get_register_value(int n)
{
	const static int reg_map[] = {
#if defined(__x86_64__)
		[UNW_X86_64_RAX] = offsetof(struct user_regs_struct, rax) / sizeof(long),
		[UNW_X86_64_RDX] = offsetof(struct user_regs_struct, rdx) / sizeof(long),
		[UNW_X86_64_RCX] = offsetof(struct user_regs_struct, rcx) / sizeof(long),
		[UNW_X86_64_RBX] = offsetof(struct user_regs_struct, rbx) / sizeof(long),
		[UNW_X86_64_RSI] = offsetof(struct user_regs_struct, rsi) / sizeof(long),
		[UNW_X86_64_RDI] = offsetof(struct user_regs_struct, rdi) / sizeof(long),
		[UNW_X86_64_RBP] = offsetof(struct user_regs_struct, rbp) / sizeof(long),
		[UNW_X86_64_RSP] = offsetof(struct user_regs_struct, rsp) / sizeof(long),
		[UNW_X86_64_RIP] = offsetof(struct user_regs_struct, rip) / sizeof(long),
#else
		[UNW_X86_EAX] = offsetof(struct user_regs_struct, eax) / sizeof(long),
		[UNW_X86_EDX] = offsetof(struct user_regs_struct, edx) / sizeof(long),
		[UNW_X86_ECX] = offsetof(struct user_regs_struct, ecx) / sizeof(long),
		[UNW_X86_EBX] = offsetof(struct user_regs_struct, ebx) / sizeof(long),
		[UNW_X86_ESI] = offsetof(struct user_regs_struct, esi) / sizeof(long),
		[UNW_X86_EDI] = offsetof(struct user_regs_struct, edi) / sizeof(long),
		[UNW_X86_EBP] = offsetof(struct user_regs_struct, ebp) / sizeof(long),
		[UNW_X86_ESP] = offsetof(struct user_regs_struct, esp) / sizeof(long),
		[UNW_X86_EIP] = offsetof(struct user_regs_struct, eip) / sizeof(long),
		[UNW_X86_EFLAGS] = offsetof(struct user_regs_struct, eflags) / sizeof(long),
		[UNW_X86_TRAPNO] = offsetof(struct user_regs_struct, orig_eax) / sizeof(long),
#endif
	};
	n = reg_map[n];
	return ((elf_greg_t*)&g_registers)[n];
}

void _crash_stack_print_regs(FILE* outputfile)
{
	fprintf(outputfile, "\nRegister Information\n");
#if defined(__x86_64__)
	fprintf(outputfile, "gs  = 0x%016llx, fs  = 0x%016llx\nes  = 0x%016llx, ds  = 0x%016llx\n",
		g_registers.gs,
		g_registers.fs,
		g_registers.es,
		g_registers.ds);
	fprintf(outputfile, "rdi = 0x%016llx, rsi = 0x%016llx\nrbp = 0x%016llx, rsp = 0x%016llx\n",
		g_registers.rdi,
		g_registers.rsi,
		g_registers.rbp,
		g_registers.rsp);
	fprintf(outputfile, "rax = 0x%016llx, rbx = 0x%016llx\nrcx = 0x%016llx, rdx = 0x%016llx\n",
		g_registers.rax,
		g_registers.rbx,
		g_registers.rcx,
		g_registers.rdx);
	fprintf(outputfile, "r8  = 0x%016llx, r9  = 0x%016llx\nr10 = 0x%016llx, r11 = 0x%016llx\n",
		g_registers.r8,
		g_registers.r9,
		g_registers.r10,
		g_registers.r11);
	fprintf(outputfile, "r12 = 0x%016llx, r13 = 0x%016llx\nr14 = 0x%016llx, r15 = 0x%016llx\n",
		g_registers.r12,
		g_registers.r13,
		g_registers.r14,
		g_registers.r15);
	fprintf(outputfile, "rip = 0x%016llx\n",
		g_registers.rip);
#else
	fprintf(outputfile, "gs  = 0x%08lx, fs  = 0x%08lx\nes  = 0x%08lx, ds  = 0x%08lx\n",
	       g_registers.xgs,
	       g_registers.xfs,
	       g_registers.xes,
	       g_registers.xds);
	fprintf(outputfile, "edi = 0x%08lx, esi = 0x%08lx\nebp = 0x%08lx, esp = 0x%08lx\n",
	       g_registers.edi,
	       g_registers.esi,
	       g_registers.ebp,
	       g_registers.esp);
	fprintf(outputfile, "eax = 0x%08lx, ebx = 0x%08lx\necx = 0x%08lx, edx = 0x%08lx\n",
	       g_registers.eax,
	       g_registers.ebx,
	       g_registers.ecx,
	       g_registers.edx);
	fprintf(outputfile, "eip = 0x%08lx\n",
	       g_registers.eip);
#endif
}
